/*
  This code is released under the Antivirial License.

  http://www.ecstaticlyrics.com/antiviral/

  The only condition of the license is that you may not change the license
  terms.  Because of this, you may *not* include this code in GPL-licensed
  projects, as the GPL requires that the entire project be GPL-licensed.
*/

#ifndef BUFFER_H
#define BUFFER_H

#include "thread.h"

struct buffer_data {
  void *memory_pointer;
  unsigned int memory_size;
  void *pointer;
  unsigned int size;
  MUTEX mutex;              // not initialized by buffer_init
};
typedef struct buffer_data BUFFER;

void buffer__init(BUFFER *);
void buffer__kill(BUFFER *);
void *buffer__write(BUFFER *, unsigned int, char *, int);
void *buffer__allocate(BUFFER *, unsigned int, char *, int);
void buffer__free(BUFFER *, unsigned int, char *, int);

#define buffer_init(pointer) buffer__init(&pointer)
#define buffer_kill(pointer) buffer__kill(&pointer)
#define buffer_write(pointer, size) buffer__write(&pointer, size, __FILE__, __LINE__)
#define buffer_allocate(pointer, size) buffer__allocate(&pointer, size, __FILE__, __LINE__)
#define buffer_free(pointer, size) buffer__free(&pointer, size, __FILE__, __LINE__)

#endif
