/*
  This code is released under the Antivirial License.

  http://www.ecstaticlyrics.com/antiviral/

  The only condition of the license is that you may not change the license
  terms.  Because of this, you may *not* include this code in GPL-licensed
  projects, as the GPL requires that the entire project be GPL-licensed.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <GLFW/glfw3.h>

#include "easy.h"
#include "memory.h"

#include "glfw.h"

static void error_callback (int code, char *text) {
  fprintf(stderr, "GLFW Error %d: %s\n", code, text);
};

static MUTEX message;
static int message_action = 0;
#define ACTION_RESPONSE 1
#define ACTION_OPEN_WINDOW 2
#define ACTION_CLOSE_WINDOW 3
static int message_window = 0;
static int message_int[16];
static void *message_void[16];

static volatile int open_window_count = 0;
static volatile int all_windows_exit_flag = 0;

static void callback_close (GLFWwindow *window) {
  struct window_data *data = glfwGetWindowUserPointer(window);
  data->close = 1;
};

static void callback_resize (GLFWwindow *window, int width, int height) {
  struct window_data *data = glfwGetWindowUserPointer(window);
  data->width = width;
  data->height = height;
};

static void callback_position (GLFWwindow *window, int x, int y) {
  struct window_data *data = glfwGetWindowUserPointer(window);
  data->window_x = x;
  data->window_y = y;
};

static void callback_iconify (GLFWwindow *window, int iconified) {
  struct window_data *data = glfwGetWindowUserPointer(window);
  if (iconified) {
    data->visible = 0;
  } else {
    data->visible = 1;
  };
};

static void callback_focus (GLFWwindow *window, int focused) {
  struct window_data *data = glfwGetWindowUserPointer(window);
  if (focused) {
    data->focus = 1;
  } else {
    data->focus = 0;
  };
};

static void callback_refresh (GLFWwindow *window) {
  struct window_data *data = glfwGetWindowUserPointer(window);
  data->refresh = 1;
};

static void callback_key (GLFWwindow *window, int key, int scancode, int action, int mods) {
  struct window_data *data = glfwGetWindowUserPointer(window);
  MUTEX_LOCK(data->mutex);
  struct event_data *event = buffer_write(data->event, sizeof(struct event_data));
  event->type = EVENT_KEY;
  event->key = key;
  event->action = action;
  event->x = data->mouse_x;
  event->y = data->mouse_y;
  event->modifiers = mods;
  MUTEX_UNLOCK(data->mutex);
};

static void callback_character (GLFWwindow* window, unsigned int codepoint) {
  //fprintf(stderr, "Character callback recevied %d a.k.a. 0x%08x\n", codepoint, codepoint);
  struct window_data *data = glfwGetWindowUserPointer(window);
  MUTEX_LOCK(data->mutex);
  struct event_data *event = buffer_write(data->event, sizeof(struct event_data));
  event->type = EVENT_CHARACTER;
  event->character = codepoint;
  event->x = data->mouse_x;
  event->y = data->mouse_y;
  event->modifiers = 0;
  MUTEX_UNLOCK(data->mutex);
};

static void callback_mouse_position (GLFWwindow* window, double xpos, double ypos) {
  struct window_data *data = glfwGetWindowUserPointer(window);
  MUTEX_LOCK(data->mutex);
  struct event_data *event = buffer_write(data->event, sizeof(struct event_data));
  event->type = EVENT_MOVE;
  event->x = data->mouse_x = xpos;
  event->y = data->mouse_y = data->height - ypos - 1;
  event->modifiers = 0;
  MUTEX_UNLOCK(data->mutex);
};

static void callback_cursor_enter (GLFWwindow* window, int entered) {
  struct window_data *data = glfwGetWindowUserPointer(window);
  MUTEX_LOCK(data->mutex);
  struct event_data *event = buffer_write(data->event, sizeof(struct event_data));
  event->type = EVENT_CLIP;
  event->action = entered ? 0 : 1;
  event->x = data->mouse_x;
  event->y = data->mouse_y;
  event->modifiers = 0;
  MUTEX_UNLOCK(data->mutex);
};

static void callback_mouse_button (GLFWwindow* window, int button, int action, int mods) {
  struct window_data *data = glfwGetWindowUserPointer(window);
  MUTEX_LOCK(data->mutex);
  struct event_data *event = buffer_write(data->event, sizeof(struct event_data));
  event->type = EVENT_BUTTON;
  event->button = button;
  event->action = action;
  event->x = data->mouse_x;
  event->y = data->mouse_y;
  event->modifiers = mods;
  MUTEX_UNLOCK(data->mutex);
};

static void callback_mouse_scroll (GLFWwindow* window, double xoffset, double yoffset) {
  struct window_data *data = glfwGetWindowUserPointer(window);
  MUTEX_LOCK(data->mutex);
  struct event_data *event = buffer_write(data->event, sizeof(struct event_data));
  event->type = EVENT_SCROLL;
  event->dx = xoffset;
  event->dy = yoffset;
  event->x = data->mouse_x;
  event->y = data->mouse_y;
  event->modifiers = 0;
  MUTEX_UNLOCK(data->mutex);
};

static volatile int thread_run_flag = 0;
static volatile int thread_exit_flag = 0;
static volatile int initialized = 0;

static void *main_thread(void *argument) {

  int result;

  glfwSetErrorCallback((GLFWerrorfun) error_callback);

  int major, minor, revision;
  glfwGetVersion(&major, &minor, &revision);
  fprintf(stderr, "GLFW shared library is version %i.%i.%i, ", major, minor, revision);
  fprintf(stderr, "accessed via headers for GLFW %i.%i.%i.\n", GLFW_VERSION_MAJOR, GLFW_VERSION_MINOR, GLFW_VERSION_REVISION);
  fprintf(stderr, "GLFW version string: \"%s\"\n", glfwGetVersionString());

  result = glfwInit();
  if (!result) easy_fuck("glfwInit() returned a non-zero value!");

  glfwWindowHint(GLFW_VISIBLE, GL_FALSE);
  GLFWwindow *share = glfwCreateWindow(64, 64, "...", NULL, NULL);
  glfwMakeContextCurrent(share);

  #include "gui.h"
  gui_initialize();

  initialized = 1;

  while (!thread_exit_flag) {
    MUTEX_LOCK(message);
    if (message_action == ACTION_OPEN_WINDOW) {
      struct window_data *data = NULL;
      memory_allocate(data, sizeof(struct window_data));
      memset(data, 0, sizeof(struct window_data));
      MUTEX_INIT(data->mutex);
      buffer_init(data->event);
      GLFWwindow *window;
      glfwWindowHint(GLFW_VISIBLE, GL_TRUE);
      glfwWindowHint(GLFW_RESIZABLE, (message_int[4] & 1) ? GL_TRUE : GL_FALSE);
      window = glfwCreateWindow(message_int[0], message_int[1], message_void[0], NULL, share);
      glfwSetWindowPos(window, message_int[2], message_int[3]);
      glfwSetWindowCloseCallback(window, callback_close);
      glfwSetFramebufferSizeCallback(window, callback_resize);
      glfwSetWindowIconifyCallback(window, callback_iconify);
      glfwSetWindowFocusCallback(window, callback_focus);
      glfwSetWindowPosCallback(window, callback_position);
      glfwSetWindowRefreshCallback(window, callback_refresh);
      glfwSetKeyCallback(window, callback_key);
      glfwSetCharCallback(window, callback_character);
      glfwSetCursorPosCallback(window, callback_mouse_position);
      glfwSetCursorEnterCallback(window, callback_cursor_enter);
      glfwSetMouseButtonCallback(window, callback_mouse_button);
      glfwSetScrollCallback(window, callback_mouse_scroll);
      glfwSetWindowUserPointer(window, data);
      memory_allocate(data->title, strlen(message_void[0]) + 1);
      strcpy(data->title, message_void[0]);
      data->window = window;
      data->refresh = 1;
      glfwGetFramebufferSize(window, &data->width, &data->height);
      message_void[0] = data;
      message_action = ACTION_RESPONSE;
    };
    if (message_action == ACTION_CLOSE_WINDOW) {
      struct window_data *data = message_void[0];
      GLFWwindow *window = data->window;
      glfwGetWindowPos(window, &message_int[0], &message_int[1]);
      glfwDestroyWindow(window);
      buffer_kill(data->event);
      memory_allocate(data->title, 0);
      memory_allocate(data, 0);
      message_action = ACTION_RESPONSE;
    };
    MUTEX_UNLOCK(message);
    glfwWaitEvents();
  };

  gui_terminate();

  glfwDestroyWindow(share);

  glfwTerminate();

  initialized = 0;
  thread_run_flag = 0;
  thread_exit();

};

static int str2int (const char *string) {
  #include "sha1.h"
  char sum[20];
  sha1(sum, string, strlen(string));
  int key = *((int *) sum);
  //printf("Key %s is %d\n", string, key);
  return key;
};

struct window_data *glfw_open_window(int width, int height, char *title, int flags) {

  open_window_count++;

  int key = str2int(title);
  //printf("Key \"%s\" is %d.\n", title, key);

  int x = 64;
  int y = 64;

  FILE *file = fopen("window_positions", "r");
  if (file) {
    int array[3];
    while (fread(array, 3 * sizeof(int), 1, file)) {
      if (array[0] == key) {
        x = array[1];
        y = array[2];
        break;
      };
    };
    fclose(file);
  };

  while (1) {
    MUTEX_LOCK(message);
    if (!message_action) {
      message_action = ACTION_OPEN_WINDOW;
      message_int[0] = width;
      message_int[1] = height;
      message_int[2] = x;
      message_int[3] = y;
      message_int[4] = flags;
      message_void[0] = title;
      MUTEX_UNLOCK(message);
      break;
    };
    MUTEX_UNLOCK(message);
    easy_sleep(0.001);
  };

  glfwPostEmptyEvent();

  struct window_data *data = NULL;

  while (1) {
    MUTEX_LOCK(message);
    if (message_action == ACTION_RESPONSE) {
      data = message_void[0];
      message_action = 0;
      MUTEX_UNLOCK(message);
      break;
    };
    MUTEX_UNLOCK(message);
    easy_sleep(0.001);
    static int kick_in_the_nuts = 0;
    if (++kick_in_the_nuts > 100) {
      fprintf(stderr, "Kick in the nuts!\n");
      glfwPostEmptyEvent();
      kick_in_the_nuts = 0;
    };
  };

  glfwMakeContextCurrent(data->window);
  glfwSwapInterval(1);

  return data;

};

void glfw_close_window(struct window_data *data) {

  int key = str2int(data->title);
  //printf("Window title \"%s\" is key %d\n", data->title, key);

  int x = 0;
  int y = 0;

  while (1) {
    MUTEX_LOCK(message);
    if (!message_action) {
      message_action = ACTION_CLOSE_WINDOW;
      message_void[0] = data;
      MUTEX_UNLOCK(message);
      break;
    };
    MUTEX_UNLOCK(message);
    easy_sleep(0.001);
  };

  glfwPostEmptyEvent();

  while (1) {
    MUTEX_LOCK(message);
    if (message_action == ACTION_RESPONSE) {
      x = message_int[0];
      y = message_int[1];
      message_action = 0;
      MUTEX_UNLOCK(message);
      break;
    };
    MUTEX_UNLOCK(message);
    easy_sleep(0.001);
    static int kick_in_the_nuts = 0;
    if (++kick_in_the_nuts > 100) {
      fprintf(stderr, "Kick in the nuts!\n");
      glfwPostEmptyEvent();
      kick_in_the_nuts = 0;
    };
  };

  //printf("Recording that window %d was at (%d, %d) when closed.\n", key, x, y);

  int fuck = open("window_positions", O_RDWR | O_CREAT, 0666);
  if (fuck >= 0) {
    FILE *file = fdopen(fuck, "r+");
    if (file) {
      fseek(file, 0, SEEK_SET);
      int array[3];
      while (fread(array, 3 * sizeof(int), 1, file)) {
        if (array[0] == key) {
          array[1] = x;
          array[2] = y;
          fseek(file, -3 * sizeof(int), SEEK_CUR);
          fwrite(array, 3 * sizeof(int), 1, file);
          break;
        };
      };
      if (feof(file)) {
        array[0] = key;
        array[1] = x;
        array[2] = y;
        fwrite(array, 3 * sizeof(int), 1, file);
      };
      fclose(file);
    } else {
      close(fuck);
    };
  };

  open_window_count--;

};

void glfw_swap_buffers (struct window_data *data) {
  glfwSwapBuffers(data->window);
};

void glfw_initialize () {

  MUTEX_INIT(message);

  initialized = 0;
  all_windows_exit_flag = 0;

  thread_run_flag = 1;
  thread_exit_flag = 0;
  thread_create(main_thread, NULL);

  while (!initialized) easy_sleep(0.001);

};

void glfw_terminate () {

  all_windows_exit_flag = 1;
  while (open_window_count) easy_sleep(0.01);

  if (thread_run_flag) {
    thread_exit_flag = 1;
    glfwPostEmptyEvent();
    while (thread_run_flag) {
      easy_sleep(0.01);
      static int kick_in_the_nuts = 0;
      if (++kick_in_the_nuts > 10) {
        fprintf(stderr, "Kick in the nuts!\n");
        glfwPostEmptyEvent();
        kick_in_the_nuts = 0;
      };
    };
  };

};

void glfw_generic_event_loop (int window_width, int window_height, char *window_title, int flags, volatile int *run_flag, void (*callback)(void *, struct window_data *, struct event_data *), void *parameter) {
  struct window_data *window;
  window = glfw_open_window(window_width, window_height, window_title, flags);
  while (!window->close && !all_windows_exit_flag) {
    glViewport(0, 0, window->width, window->height);
    glMatrixMode(GL_MODELVIEW); glLoadIdentity();
    glTranslated(-1.0, -1.0, 0.0);
    glScaled(2.0 / window->width, 2.0 / window->height, 1.0);
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    struct event_data event;
    do {
      memset(&event, 0, sizeof(struct event_data));
      MUTEX_LOCK(window->mutex);
      if (window->event.size >= sizeof(struct event_data)) {
        memmove(&event, window->event.pointer, sizeof(struct event_data));
        buffer_free(window->event, sizeof(struct event_data));
      };
      MUTEX_UNLOCK(window->mutex);
      if (event.type) {
        callback(parameter, window, &event);
      } else {
        callback(parameter, window, NULL);
      };
    } while (event.type);
    glfw_swap_buffers(window);
  };
  glfw_close_window(window);
  if (run_flag != NULL) *run_flag = 0;
};

struct event_loop_thread_parameters {
  int window_width;
  int window_height;
  char *window_title;
  int flags;
  volatile int *run_flag;
  void (*callback)(void *, struct window_data *, struct event_data *);
  void *parameter;
};

static void *event_loop_thread (void *parameter) {
  struct event_loop_thread_parameters *parameters = parameter;
  glfw_generic_event_loop(parameters->window_width, parameters->window_height, parameters->window_title, parameters->flags, parameters->run_flag, parameters->callback, parameters->parameter);
  memory_allocate(parameters, 0);
};

void glfw_generic_event_loop_thread (int window_width, int window_height, char *window_title, int flags, volatile int *run_flag, void (*callback)(void *, struct window_data *, struct event_data *), void *parameter) {

  struct event_loop_thread_parameters *parameters = NULL;

  memory_allocate(parameters, sizeof(struct event_loop_thread_parameters));

  parameters->window_width = window_width;
  parameters->window_height = window_height;
  parameters->window_title = window_title;
  parameters->flags = flags;
  parameters->run_flag = run_flag;
  parameters->callback = callback;
  parameters->parameter = parameter;

  thread_create(event_loop_thread, parameters);

};
