/*
  This code is released under the Antivirial License.

  http://www.ecstaticlyrics.com/antiviral/

  The only condition of the license is that you may not change the license
  terms.  Because of this, you may *not* include this code in GPL-licensed
  projects, as the GPL requires that the entire project be GPL-licensed.
*/

#ifndef MOOSFET_GLFW_H
#define MOOSFET_GLFW_H

#include <GLFW/glfw3.h>
#include "thread.h"
#include "buffer.h"

struct window_data {
  GLFWwindow *window;
  char *title;
  int close;
  int refresh;
  int visible;
  int focus;
  int width;
  int height;
  int window_x;
  int window_y;
  int mouse_x;
  int mouse_y;
  MUTEX mutex;
  BUFFER event;
};

#define EVENT_NONE 0
#define EVENT_KEY 1
#define EVENT_CHARACTER 2
#define EVENT_BUTTON 3
#define EVENT_SCROLL 4
#define EVENT_MOVE 5
#define EVENT_CLIP 6

struct event_data {
  int type;
  union {
    unsigned int character;
    int key;
    int button;
    int dx;
  };
  union {
    int action;
    int dy;
  };
  int modifiers;
  int x;
  int y;
};

void glfw_initialize();
void glfw_terminate();
struct window_data *glfw_open_window(int width, int height, char *title, int flags);
void glfw_close_window(struct window_data *data);
void glfw_swap_buffers(struct window_data *data);

void glfw_generic_event_loop (int window_width, int window_height, char *window_title, int flags, volatile int *run_flag, void (*callback)(void *, struct window_data *, struct event_data *), void *parameter);
void glfw_generic_event_loop_thread (int window_width, int window_height, char *window_title, int flags, volatile int *run_flag, void (*callback)(void *, struct window_data *, struct event_data *), void *parameter);

#endif
