#include <stdlib.h>
#include <stdio.h>
#include <GL/gl.h>
#include <string.h>
#include <stdarg.h>

#include "memory.h"
#include "glfw.h"
#include "easy.h"

#include "font.h"

static GLuint texture_font;

void gui_initialize () {
  glGenTextures(1, &texture_font);
  glBindTexture(GL_TEXTURE_2D, texture_font);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, font_width, font_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, &font_image);
};

void gui_terminate() {
  glDeleteTextures(1, &texture_font);
};

void gui_text(struct event_data *event, int x, int y, char *text) {
  if (!event) {
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, texture_font);
    //glEnable(GL_BLEND); glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_ALPHA_TEST); glAlphaFunc(GL_GEQUAL, 0.5f);
    glBegin(GL_QUADS);
    for (int i = 0; i < strlen(text); i++) {
      int c = text[i];
      if (c < 32 || c > 126) c = 127;
      int cx = c % 16;
      int cy = c / 16;
      glTexCoord2d((cx + 0) / 16.0, (cy + 0) / -8.0); glVertex2f(8 * i + x, y);
      glTexCoord2d((cx + 1) / 16.0, (cy + 0) / -8.0); glVertex2f(8 * i + x + 8, y);
      glTexCoord2d((cx + 1) / 16.0, (cy + 1) / -8.0); glVertex2f(8 * i + x + 8, y + 16);
      glTexCoord2d((cx + 0) / 16.0, (cy + 1) / -8.0); glVertex2f(8 * i + x, y + 16);
    };
    glEnd();
    //glDisable(GL_BLEND);
    glDisable(GL_ALPHA_TEST);
    glDisable(GL_TEXTURE_2D);
  };
};

void gui_printf(struct event_data *event, int x, int y, const char *format, ...) {
  if (!event) {
    va_list ooo;
    va_start(ooo, format);
    int count = vsnprintf(NULL, 0, format, ooo);
    va_end(ooo);
    count++;
    char *text = NULL;
    memory_allocate(text, count);
    va_start(ooo, format);
    vsnprintf(text, count, format, ooo);
    va_end(ooo);
    gui_text(event, x, y, text);
    memory_allocate(text, 0);
  };
};

#define MAX_WIDGETS 1024

struct widget_data {
  int over;
  int click;
};

__thread struct widget_data widget[MAX_WIDGETS];

#include "color.h"

int gui_radio(struct event_data *event, int index, int x, int y, char *text, int *variable, int value) {

  int c; // color
  if (variable == NULL) {
    c = 6;
  } else {
    if (*variable == value) {
      c = 5;
    } else {
      c = 1;
    };
  };

  if (index >= MAX_WIDGETS) easy_fuck("too many widgets in a window");

  int result = 0;

  int x1 = x - 4;
  int x2 = x + 8 * strlen(text) + 4;
  int y1 = y;
  int y2 = y + 16;

  if (event) {
    if (event->type == EVENT_MOVE) {
      if (x1 <= event->x && event->x < x2 && y1 <= event->y && event->y < y2) {
        widget[index].over = 1;
      } else {
        widget[index].over = 0;
      };
    } else if (event->type == EVENT_CLIP) {
      widget[index].over = 0;
    } else if (event->type == EVENT_BUTTON && event->button == 0) {
      if (event->action == GLFW_PRESS) {
        if (x1 <= event->x && event->x < x2 && y1 <= event->y && event->y < y2) {
          widget[index].click = 1;
        };
      } else if (event->action == GLFW_RELEASE && widget[index].click) {
        widget[index].click = 0;
        if (x1 <= event->x && event->x < x2 && y1 <= event->y && event->y < y2) {
          result = 1;
        };
      };
    };
  } else {
    float highlight = 0.0f;
    if (widget[index].over) {
      if (widget[index].click) {
        highlight = 0.25f;
      } else {
        highlight = 0.25f;
      };
    } else {
      highlight = 0.0f;
    };
    glBegin(GL_QUADS);
    if (widget[index].over) {
      if (widget[index].click) {
        //glColor4f(0.2, 0.2, 0.2, 0.5);
        color(c, -5);
      } else {
        //glColor4f(0.2, 0.2, 0.2, 0.5);
        color(c, -5);
      };
    } else {
      //glColor4f(0.0, 0.0, 0.0, 0.5);
      color(c, -9);
    };
    glVertex2f(x1 - 0, y1);
    glVertex2f(x2 + 0, y1);
    if (widget[index].over) {
      if (widget[index].click) {
        //glColor4f(0.0, 0.0, 0.0, 0.5);
        color(c, -9);
      } else {
        //glColor4f(0.7, 0.7, 0.7, 0.5);
        color(c, +5);
      };
    } else {
      //glColor4f(0.5, 0.5, 0.5, 0.5);
      color(c, 0);
    };
    glVertex2f(x2 + 0, y2);
    glVertex2f(x1 - 0, y2);
    glEnd();
    glColor3f(1.0, 1.0, 1.0);
    gui_text(event, x, y, text);
  };

  if (result && variable != NULL) {
    *variable = value;
  };

  return result;

};

int gui_slider(struct event_data *event, int index, int x, int y, int width, int height, float min, float max, volatile float *variable) {

  int change = 0;

  if (event) {
    int x1 = x;
    int x2 = x + width;
    int y1 = y;
    int y2 = y + height;
    int modify = 0;
    if (event->type == EVENT_MOVE) {
      if (x1 <= event->x && event->x < x2 && y1 <= event->y && event->y < y2) {
        widget[index].over = 1;
      } else {
        widget[index].over = 0;
      };
      if (widget[index].click) modify = 1;
    } else if (event->type == EVENT_CLIP) {
      widget[index].over = 0;
    } else if (event->type == EVENT_BUTTON && event->button == 0) {
      if (event->action == GLFW_PRESS) {
        if (x1 <= event->x && event->x < x2 && y1 <= event->y && event->y < y2) {
          widget[index].click = 1;
          modify = 1;
        };
      } else if (event->action == GLFW_RELEASE && widget[index].click) {
        widget[index].click = 0;
      };
    };
    if (modify) {
      float new;
      if (height > width) {
        new = (max - min) * (event->y - y - 0.5f) / (height - 2) + min;
      } else {
        new = (max - min) * (event->x - x - 0.5f) / (width - 2) + min;
      };
      if (new < min) new = min;
      if (new > max) new = max;
      if (*variable != new) {
        *variable = new;
        change = 1;
      };
    };
  } else {

    glBegin(GL_QUADS);

    if (widget[index].over || widget[index].click) {
      color(5, 0);
    } else {
      color(1, 0);
    };
    glVertex2f(x, y);
    glVertex2f(x + width, y);
    glVertex2f(x + width, y + height);
    glVertex2f(x, y + height);

    x++; y++;
    width -= 2;
    height -= 2;

    color(8, -9);
    glVertex2f(x, y);
    glVertex2f(x + width, y);
    glVertex2f(x + width, y + height);
    glVertex2f(x, y + height);

    if (height > width) {
      float v = height * (*variable - min) / (max - min);
      float m = height * (0.0f - min) / (max - min);
      if (min >= 0.0f) m = 0.0f;
      color(8, -9);
      glVertex2f(x, y + m);
      glVertex2f(x + width, y + m);
      color(5, 0);
      glVertex2f(x + width, y + v);
      glVertex2f(x, y + v);
    } else {
      float v = width * (*variable - min) / (max - min);
      float m = width * (0.0f - min) / (max - min);
      if (min >= 0.0f) m = 0.0f;
      color(8, -9);
      glVertex2f(x + m, y + height);
      glVertex2f(x + m, y);
      color(5, 0);
      glVertex2f(x + v, y);
      glVertex2f(x + v, y + height);
    };

    glEnd();

  };

  return change;

};
