/*
  This code is released under the Antivirial License.

  http://www.ecstaticlyrics.com/antiviral/

  The only condition of the license is that you may not change the license
  terms.  Because of this, you may *not* include this code in GPL-licensed
  projects, as the GPL requires that the entire project be GPL-licensed.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <complex.h>
#include <locale.h>
#include <arpa/inet.h>

#include <sys/types.h>  // for open()
#include <sys/stat.h>   // for open()
#include <fcntl.h>      // for open()
#include <sys/ioctl.h>  // for ioctl()
#include <unistd.h>     // for read()

#include <sys/stat.h>   // for mkdir()
#include <sys/types.h>  // for mkdir()

#include "easy.h"
#include "thread.h"
#include "buffer.h"
#include "fft.h"
#include "user.h"
#include "alsa.h"
#include "radio.h"
#include "memory.h"
#include "option.h"
#include "config.h"
#include "common.h"
#include "filter.h"
#include "network.h"

volatile int radio_test_flag = 0;
volatile int iq_playback_offset = 0;
static volatile int set_squelch_flag;
static float squelch_level = 0.000001f;
volatile float radio_signal_strength_squared = 0.0f;
volatile float radio_iq_strength_squared = 0.0f;
volatile float radio_audio_gain = -36.0f;
volatile int radio_scan_enable = 0;
volatile int radio_invert_audio = 0;

volatile int radio_channel_bandwidth = 10;
volatile int radio_audio_bandwidth = 5;

static volatile int thread_run_flag = 0;
static volatile int thread_exit_flag = 0;
static volatile int initialized = 0;

MUTEX iq_buffer_mutex;
char *iq_buffer_pointer = NULL;
int iq_buffer_offset = 0;
int iq_buffer_size = 0;

MUTEX audio_fft_buffer_mutex;
BUFFER audio_fft_buffer;

MUTEX signal_fft_buffer_mutex;
BUFFER signal_fft_buffer;

static volatile int record_audio_flag = 0;
volatile int radio_demodulator_enable = 0;

static int record_iq_flag = 0;

#define PP_CAT(a, b) PP_CAT_I(a, b)
#define PP_CAT_I(a, b) PP_CAT_II(~, a ## b)
#define PP_CAT_II(p, res) res
#define UNIQUE_NAME(base) PP_CAT(base, __LINE__)

#define ONCE_EVERY_(n) static int UNIQUE_NAME(N); if (++UNIQUE_NAME(N) >= n) UNIQUE_NAME(N) = 0; if (UNIQUE_NAME(N) == 0)

volatile struct radio_state radio_state = {};

#if FILTER_QUALITY == 0
  #include "better_fir_filters.c"
#elif FILTER_QUALITY == 1
  #include "ok_fir_filters.c"
#elif FILTER_QUALITY == 2
  #include "worse_fir_filters.c"
#else
  #error FILTER_QUALITY must be 0, 1 or 2.
#endif

static void send_packet (int id, int v) {
  char packet[5];
  packet[0] =id;
  *((int *) &packet[1]) = htonl(v);
  //fwrite(packet, 5, 1, stdout); fflush(stdout);
  int count = 5, written = 0;
  while (written < count) {
    int result = write(network_socket, packet + written, count - written);
    if (result < 0) {
      fprintf(stderr, "Network error when sending command to rtl_tcp.\n");
      exit(1);
    };
    written += result;
  };
};

void radio_set_squelch () {
  set_squelch_flag = 1;
};

void radio_start_recording () {
  record_audio_flag = 1;
};

void radio_stop_recording () {
  record_audio_flag = 0;
};

void radio_start_iq_recording () {
  record_iq_flag = 1;
};

void radio_set_time_offset (int y) {
  int offset = y * 2 * WATERFALL_FFT_SIZE * (1 << option_waterfall_slow);
  if (offset < 0) offset = 0;
  MUTEX_LOCK(iq_buffer_mutex);
  iq_playback_offset = iq_buffer_offset - offset;
  if (iq_playback_offset < 0) iq_playback_offset += iq_buffer_size;
  iq_playback_offset &= ~1;
  MUTEX_UNLOCK(iq_buffer_mutex);
};

void radio_get_playback_position(float *x, float *y) {
  *x = (radio_state.offset_frequency + (IQ_SAMPLE_RATE / 1024.0f / 2.0f)) / (float) radio_state.sample_rate + 0.5;
  MUTEX_LOCK(iq_buffer_mutex);
  int offset = iq_buffer_offset - iq_playback_offset;
  if (offset < 0) offset += iq_buffer_size;
  MUTEX_UNLOCK(iq_buffer_mutex);
  *y = offset / (2.0f * WATERFALL_FFT_SIZE * (1 << option_waterfall_slow));
};

void radio_demodulate (float *output, int count) {

  #define BUNCH (IQ_SAMPLE_RATE / 48000)

  char *data = NULL;
  int bytes_wanted = 2 * count * BUNCH;
  memory_allocate(data, bytes_wanted);

  // grab data from the ring buffer

  MUTEX_LOCK(iq_buffer_mutex);
  int available = iq_buffer_offset - iq_playback_offset;
  if (available < 0) available += iq_buffer_size;
  if (available >= bytes_wanted) {
    // There are enough bytes, but can we read them all at once?
    if (iq_buffer_size - iq_playback_offset >= bytes_wanted) {
      memmove(data, iq_buffer_pointer + iq_playback_offset, 2 * count * BUNCH);
      iq_playback_offset += 2 * count * BUNCH;
    } else {
      int c = iq_buffer_size - iq_playback_offset;
      int d = bytes_wanted - c;
      memmove(data, iq_buffer_pointer + iq_playback_offset, c);
      memmove(data + c, iq_buffer_pointer, d);
      iq_playback_offset = d;
    };
  } else {
    available = 0;
  };
  MUTEX_UNLOCK(iq_buffer_mutex);

  // If not enough bytes, or demodulator disabled, return silence.

  if (!available || !radio_demodulator_enable) {
    memory_allocate(data, 0);
    memset(output, 0, 2 * count * sizeof(float));
    return;
  };

  float *audio_copy = NULL;
  memory_allocate(audio_copy, count * sizeof(float));

  complex float *signal_copy = NULL;
  memory_allocate(signal_copy, count * sizeof(complex float));

  int mode = radio_state.mode;

  // Process the IQ data!

  for (int i = 0; i < count * BUNCH; i++) {

    // This counter tells us when to compensate for cumulative rounding errors.
    static int check_counter = 0;
    if (++check_counter >= IQ_SAMPLE_RATE / 1000) check_counter = 0;

    float audio;
    float stereo = 0.0f;

    // Turn IQ data into a complex number:

    static double dc_i = 127.5;
    static double dc_q = 127.5;
    dc_i *= 0.999999; dc_i += 0.000001 * data[2 * i + 0];
    dc_q *= 0.999999; dc_q += 0.000001 * data[2 * i + 1];
    complex float sample = ((data[2 * i + 0] - dc_i) + (data[2 * i + 1] - dc_q) * I) / 127.5f;

    // Frequency-shift so that the frequency of interest is at 0 Hz...

    {
      static int last_of = 0;
      static complex float step = 0.0f;
      static complex float phase = 1.0f;
      static int check_counter = 0;
      if (last_of != radio_state.offset_frequency) {
        last_of = radio_state.offset_frequency;
        step = cexpf(2 * M_PI * I * -1.0 * radio_state.offset_frequency / radio_state.sample_rate);
        //fprintf(stderr, "new phase: %f, %f\n", crealf(step), cimagf(step));
      };
      phase *= step; sample *= phase;
      if (!check_counter) phase *= 1.0f / cabsf(phase);
    };

    if (mode == RADIO_MODE_FM_200) {

      // Filter the IQ down to 200 kHz.

      sample = fir_filter_2400_200 (sample, 0);

      ONCE_EVERY_(10) { // now it's 240 kHz...

        // Measure signal strength...

        float now = cabsf(sample);
        float t = radio_signal_strength_squared;
        t *= 0.99999; t += 0.00001 * now * now;
        radio_signal_strength_squared = t;

        // Demodulate...

        static complex float phase = 1.0f;
        complex float change = sample / phase;
        if (cabsf(sample) > 0.000001f) phase = cexpf(I * carg(sample));
        float deviation = 240000.0f * carg(change) / (2 * M_PI);
        float max_deviation = 75000.0;
        audio = deviation / max_deviation;

        // Broadcast FM is maximumly modulated, to the point that many of the waveforms
        // are clipped, whereas most other radio transmissions are undermodulated.
        // So to keep people from going deaf, let's quite this shit down...

        audio *= 0.2f;

        // Filter audio...

        audio = fir_filter_20_2_1 (audio, 0);
        ONCE_EVERY_(5) { // now it's 48 kHz...
          audio = fir_filter_48_15 (audio, 0);
          goto MOTHERFUCKER;
        };

        // OMG, did I really just do that?

      };

    } else {

      // Filter the IQ data down to 48 kHz...
      // FIR filters are CPU-intensive, so we're going to do this as many steps
      // which each cut the sampling rate in half, resulting in the next step
      // requiring less CPU time than previous step.  ...at least I hope.

      // filter_8_2_1 filters an e.g. 8 kHz sample rate, eliminating all frequencies above 2 kHz, and passing all below 1 kHz.
      // This means that signals up to -2 kHz through +2 kHz still gets through, a bandwidth of 4 kHz, so this filter is good
      // to cut the sample rate by 1/2, but only accurately preserves samples below 1/4 of the sample rate.

      sample = fir_filter_8_2_1 (sample, 0);
      ONCE_EVERY_(2) { // now it's 1200 kHz...

        sample = fir_filter_20_2_1 (sample, 0);
        ONCE_EVERY_(5) { // now it's 240 kHz...

          sample = fir_filter_20_2_1 (sample, 1);
          ONCE_EVERY_(5) { // now it's 48 kHz...

            // Save a copy for a close-up FFT view...

            signal_copy[i/BUNCH] = sample;

            // Filter the IQ data to the actual channel bandwidth...

            if (mode == RADIO_MODE_USB || mode == RADIO_MODE_LSB) {
              // Don't apply filter in this case.
            } else if (mode == RADIO_MODE_AM_10) {
              sample = fir_filter_48_10 (sample, 0);
            } else if (mode == RADIO_MODE_FM_12) {
              sample = fir_filter_48_12 (sample, 0);
            } else if (mode == RADIO_MODE_FM_25) {
              sample = fir_filter_48_25 (sample, 0);
            } else {
              fprintf(stderr, "No signal filter for selected demodulation mode.\n");
              exit(1);
            };

            // Detect carrier frequency offset...

            if (mode != RADIO_MODE_USB && mode != RADIO_MODE_LSB) {
              static float offset = 0.0f;
              static complex float phase = 1.0;
              complex float change = sample / phase;
              if (cabsf(sample) > 0.000001f) phase = cexpf(I * carg(sample));
              float error = cargf(change) / (2 * M_PI);
              offset = 0.9999 * offset + 0.0001 * error;
              radio_carrier_frequency_offset = 48000.0f * offset;
            };

            // Perform SSB demodulation...

            if (mode == RADIO_MODE_USB || mode == RADIO_MODE_LSB) {

              // Filter so that only the two sidebands are present...

              sample = fir_filter_48_10 (sample, 0);

              // Then up-shift or down-shift by 10 kHz...

              static float a = 0.0f;
              a += 2 * M_PI * 10000.0f / 48000.0f;
              if (a > M_PI) a -= 2 * M_PI;
              if (a < -M_PI) a += 2 * M_PI;

              int usb = mode == RADIO_MODE_LSB;

              if (usb) {
                sample *= cexpf(+a * I);
              } else {
                sample *= cexpf(-a * I);
              };

              // Then filter out the unwanted sideband...

              sample = fir_filter_ssb (sample, 0);

              // Then reverse the shift...

              if (usb) {
                sample *= cexpf(-a * I);
              } else {
                sample *= cexpf(+a * I);
              };

              // Then turn it into audio...

              audio = crealf(sample);

            };

            // Measure signal strength...

            float now = cabsf(sample);
            float t = radio_signal_strength_squared;
            t *= 0.999; t += 0.001 * now * now;
            radio_signal_strength_squared = t;

            static float strength = 0.0f;
            static float squelch = 1.0f;
            strength *= 0.99;
            strength += 0.01 * now;
            if (mode == RADIO_MODE_AM_10 || mode == RADIO_MODE_FM_12 || mode == RADIO_MODE_FM_25) {
              if (strength > squelch_level) {
                squelch *= 0.99f;
                squelch += 0.01f;
              } else {
                squelch *= 0.99f;
              };
              if (set_squelch_flag) {
                squelch_level = strength * 1.5f;
                set_squelch_flag = 0;
              };
            } else {
              squelch = 1.0f;
            };
            float squelch_db = 20.0f * logf(squelch_level) / logf(10.0f);

            // Perform AM demodulation...

            if (mode == RADIO_MODE_AM_10) {

              audio = cabsf(sample);
              audio -= strength;

              // Amplify the audio...
              float scale = strength;
              float minimum = powf(10.0f, radio_audio_gain / 20.0f);
              if (scale < minimum) scale = minimum;
              audio /= scale;

            };

            // Perform FM demodulation...

            if (mode == RADIO_MODE_FM_12 || mode == RADIO_MODE_FM_25) {

              static complex float phase = 1.0f;
              complex float change = sample / phase;
              if (cabsf(sample) > 0.000001f) phase = cexpf(I * carg(sample));
              float deviation = 48000.0f * carg(change) / (2 * M_PI);
              float max_deviation;
              if (mode == RADIO_MODE_FM_12) max_deviation = 2500;
              if (mode == RADIO_MODE_FM_25) max_deviation = 5000;
              audio = deviation / max_deviation;

              // Amplify the audio...
              float scale = strength;
              float maximum = powf(10.0f, radio_audio_gain / 20.0f);
              if (scale > maximum) scale = maximum;
              audio *= scale / maximum;

            };

            if (mode == RADIO_MODE_USB || mode == RADIO_MODE_LSB) {

              // Adjust the volume of SSB audio...

              float scale = strength;
              float minimum = powf(10.0f, radio_audio_gain / 20.0f);
              if (scale < minimum) scale = minimum;
              audio /= 2.0f * scale;

            };

            // Perform audio filtering...

            if (mode == RADIO_MODE_USB || mode == RADIO_MODE_LSB) {
              // Don't apply filter in this case.
            } else if (mode == RADIO_MODE_AM_10) {
              audio = fir_filter_48_3 (audio, 0);
            } else if (mode == RADIO_MODE_FM_12) {
              audio = fir_filter_48_3 (audio, 0);
            } else if (mode == RADIO_MODE_FM_25) {
              audio = fir_filter_48_3 (audio, 0);
            } else {
              fprintf(stderr, "No audio filter for selected demodulation mode.\n");
              exit(1);
            };

            if (radio_invert_audio) {
              static float a = 0.0f;
              a += 2 * M_PI * 3400.0f / 48000.0f;
              if (a > M_PI) a -= 2 * M_PI;
              if (a < -M_PI) a += 2 * M_PI;
              audio *= sin(a);
              audio = fir_filter_48_3 (audio, 1);
            };

            MOTHERFUCKER:

            audio *= 1.0f / sqrtf(2.0f);  // Let's leave 3 dB of headroom...

            // Write the final audio to return buffer...

            *output++ = squelch * (audio + stereo);
            *output++ = squelch * (audio - stereo);

            // ...and junk...

            audio_copy[i/BUNCH] = audio;

            // ...and maybe save it too...

            static FILE *recording_file = NULL;
            static int recording = 0;

            if (record_audio_flag) {
              if (recording_file == NULL) {
                char filename[64];
                sprintf(filename, "recordings/audio_%ld.raw", time(NULL));
                mkdir("recordings", 0666);
                fprintf(stderr, "Creating %s\n", filename);
                recording_file = fopen(filename, "wb");
                if (recording_file == NULL) {
                  record_audio_flag = 0;
                };
              } else {
                char byte = nearbyint(127.5f + 127.5f * audio);
                fwrite(&byte, 1, 1, recording_file);
              };
            } else {
              if (recording_file != NULL) {
                fprintf(stderr, "Closing audio recording...\n");
                fclose(recording_file);
                recording_file = NULL;
              };
            };

          };
        };
      };

    };

  };

  memory_allocate(data, 0);

  // Store the audio and signal for FFT processing and junk...

  void *pointer;

  MUTEX_LOCK(audio_fft_buffer_mutex);
  pointer = buffer_write(audio_fft_buffer, count * sizeof(float));
  memmove(pointer, audio_copy, count * sizeof(float));
  MUTEX_UNLOCK(audio_fft_buffer_mutex);
  memory_allocate(audio_copy, 0);

  MUTEX_LOCK(signal_fft_buffer_mutex);
  pointer = buffer_write(signal_fft_buffer, count * sizeof(complex float));
  memmove(pointer, signal_copy, count * sizeof(complex float));
  MUTEX_UNLOCK(signal_fft_buffer_mutex);
  memory_allocate(signal_copy, 0);


};

static void *thread (void *parameter) {

  //open_sdr();

  iq_buffer_size = 2 * WATERFALL_FFT_SIZE * (1 << option_waterfall_slow) * 768;
  memory_allocate(iq_buffer_pointer, iq_buffer_size);
  fprintf(stderr, "I don't know why this memset() takes so fucking long sometimes, but it is in\n");
  fprintf(stderr, "radio.c on line %d if you'd like to delete it.  The only effect will be that\n", __LINE__ + 3);
  fprintf(stderr, "you may hear garbage noise if you click into the part of the waterfall where\n");
  fprintf(stderr, "no data has been recorded yet, assuming you're the type to do such a thing.\n");
  //memset(iq_buffer_pointer, 0, iq_buffer_size);
  MUTEX_INIT(iq_buffer_mutex);

  buffer_init(signal_fft_buffer);
  MUTEX_INIT(signal_fft_buffer_mutex);

  buffer_init(audio_fft_buffer);
  MUTEX_INIT(audio_fft_buffer_mutex);

  initialized = 1;

  complex float input[WATERFALL_FFT_SIZE];
  complex float output[WATERFALL_FFT_SIZE];
  struct fft_state fft_state = {};
  fft_init(&fft_state, WATERFALL_FFT_SIZE, 0);

  struct fft_state audio_fft_state = {};
  fft_init(&audio_fft_state, 512, 0);

  struct fft_state signal_fft_state = {};
  fft_init(&signal_fft_state, 256, 0);

  char data[4096];

  static struct radio_state previous_state = {};
  memset(&previous_state, 255, sizeof(previous_state));

  while (!thread_exit_flag) {

    // Adjust radio settings if anything in radio_state has changed.

    if (radio_state.direct_sampling != previous_state.direct_sampling) {
      // When direct sampling is enabled/disabled, rtl_tcp seems to
      // forget its old gain settings, so we need to remind it.
      // To do this, we just forget that we ever told it what they were.
      memset(&previous_state, 255, sizeof(previous_state));
      previous_state.direct_sampling = radio_state.direct_sampling;
      send_packet(0x09, previous_state.direct_sampling);
    };
    if (radio_state.center_frequency != previous_state.center_frequency) {
      previous_state.center_frequency = radio_state.center_frequency;
      send_packet(0x01, previous_state.center_frequency);
    };
    if (radio_state.sample_rate != previous_state.sample_rate) {
      previous_state.sample_rate = radio_state.sample_rate;
      send_packet(0x02, previous_state.sample_rate);
    };
    if (radio_state.gain_mode != previous_state.gain_mode) {
      previous_state.gain_mode = radio_state.gain_mode;
      send_packet(0x03, previous_state.gain_mode);
      if (previous_state.gain_mode == 1) {
        previous_state.gain_index = radio_state.gain_index;
        send_packet(0x0D, previous_state.gain_index);
      };
    };
    if (previous_state.gain_mode == 1) {
      if (radio_state.gain_index != previous_state.gain_index) {
        previous_state.gain_index = radio_state.gain_index;
        send_packet(0x0D, previous_state.gain_index);
      };
    };
    if (radio_state.agc_mode != previous_state.agc_mode) {
      previous_state.agc_mode = radio_state.agc_mode;
      send_packet(0x08, previous_state.agc_mode);
    };

    // Read IQ data from stdin...

    struct timeval tv;
    tv.tv_sec = 0;
    tv.tv_usec = 10000;
    fd_set reads = {}, writes = {};
    FD_SET(network_socket, &reads);
    int return_value = select(8 * sizeof(fd_set), &reads, &writes, NULL, &tv);

    if (return_value > 0) {

      int count = read(network_socket, data, 4096);
      if (count <= 0) {
        fprintf(stderr, "ERROR: read() from SDR returned %d\n", count);
        exit(1);
      };

      // Apply some dumb as shit noise filtering...

      #if 0
      if (radio_test_flag) {
        for (int i = 0; i < count / 2; i++) {
          int x = data[2 * i + 0] - 128;
          int y = data[2 * i + 1] - 128;
          static float peak = 0.0f;
          if (abs(x) > peak) {
            if (x < 0) x = -peak;
            if (x > 0) x = +peak;
            peak += 0.01;
          };
          if (abs(y) > peak) {
            if (y < 0) y = -peak;
            if (y > 0) y = +peak;
            peak += 0.01;
          };
          peak -= 0.0001;
          data[2 * i + 0] = x + 128;
          data[2 * i + 1] = y + 128;
          //fprintf(stderr, "%d, %d\n", x, y);
        };
      };
      #endif

      // Create fake test IQ data...

      #if 0
      for (int i = 0; i < count / 2; i++) {
        static double a = 0.0;
        a += 2 * M_PI * 16.0 / 1024.0;
        if (a > M_PI) a -= 2.0 * M_PI;
        if (a < -M_PI) a += 2.0 * M_PI;
        complex float signal = cexpf(a * I);
        int x = nearbyint(127 + 127 * crealf(signal) + easy_random(65536) / 65536.0f);
        int y = nearbyint(127 + 127 * cimagf(signal) + easy_random(65536) / 65536.0f);
        if (x < 0) x = 0; if (x > 255) x = 255;
        if (y < 0) y = 0; if (y > 255) y = 255;
        data[2 * i + 0] = x;
        data[2 * i + 1] = y;
        //fprintf(stderr, "%d, %d\n", x, y);
      };
      #endif

      // Create a test signal in the IQ data...
      #if 0
      for (int i = 0; i < count / 2; i++) {
        static FILE *test_audio = NULL;
        static short last_sample = 0.0f;
        static int sample_repeat_counter = 0;
        if (test_audio == NULL) {
          test_audio = fopen("test_audio.raw", "rb");

        };
        if (test_audio) {
          if (sample_repeat_counter <= 0) {
            int count = fread(&last_sample, 1, 2, test_audio);
            float s = last_sample / 32768.0f;
            s = fir_filter_48_3 (s, 1);
            last_sample = s * 32768;
            sample_repeat_counter = 50;
            if (count < 1) {
              fclose(test_audio);
            };
          };
        };
        if (test_audio) {
          static double a = 0.0;
          a += 2 * M_PI * (100.0 + (2.5 / 2.0) * (last_sample / 32768.0)) / 2400.0;
          //a += 2 * M_PI * 100.0 / 2400.0;
          if (a > M_PI) a -= 2.0 * M_PI;
          if (a < -M_PI) a += 2.0 * M_PI;
          complex float signal = cexpf(a * I);
          //signal *= 0.8 * (last_sample / 65536.0) + 0.5;
          int x = nearbyint(127 + 64 * crealf(signal) + easy_random(65536) / 32768.0f);
          int y = nearbyint(127 + 64 * cimagf(signal) + easy_random(65536) / 32768.0f);
          if (x < 0) x = 0; if (x > 255) x = 255;
          if (y < 0) y = 0; if (y > 255) y = 255;
          data[2 * i + 0] = x;
          data[2 * i + 1] = y;
          sample_repeat_counter--;
        };
      };
      #endif

      // Open/close IQ recording file...

      static FILE *recording_file = NULL;
      if (record_iq_flag) {
        if (recording_file == NULL) {
          char filename[64];
          sprintf(filename, "recordings/iq_%ld.raw", time(NULL));
          mkdir("recordings", 0666);
          fprintf(stderr, "Creating %s\n", filename);
          recording_file = fopen(filename, "wb");
          if (recording_file == NULL) {
            record_iq_flag = 0;
          };
        } else {
          fwrite(data, 1, count, recording_file);
        };
      };

      // Process the IQ data...

      MUTEX_LOCK(iq_buffer_mutex);

      // Append the new data to the ring buffer...

      if (iq_buffer_offset + count < iq_buffer_size) {
        memmove(iq_buffer_pointer + iq_buffer_offset, data, count);
        iq_buffer_offset += count;
      } else {
        int c = iq_buffer_size - iq_buffer_offset;
        int d = count - c;
        memmove(iq_buffer_pointer + iq_buffer_offset, data, c);
        memmove(iq_buffer_pointer, data + c, d);
        iq_buffer_offset = d;
      };

      while (1) {

        // As long as there's enough unprocessed data for WATERFALL_FFT_SIZE,
        // pull out a new chunk to process for the waterfall.

        static int waterfall_offset = 0;
        char fft_data[2 * WATERFALL_FFT_SIZE];

        int bytes_wanted = 2 * WATERFALL_FFT_SIZE;
        if (iq_buffer_offset >= waterfall_offset + bytes_wanted || (iq_buffer_offset < waterfall_offset && waterfall_offset + bytes_wanted < iq_buffer_size)) {
          memmove(fft_data, iq_buffer_pointer + waterfall_offset, bytes_wanted);
          waterfall_offset += 2 * WATERFALL_FFT_SIZE;
          waterfall_offset -= WATERFALL_FFT_SIZE; if (waterfall_offset < 0) waterfall_offset += iq_buffer_size;
        } else if (iq_buffer_offset < waterfall_offset && (iq_buffer_size - waterfall_offset + iq_buffer_offset) >= bytes_wanted) {
          int c = iq_buffer_size - waterfall_offset;
          int d = bytes_wanted - c;
          memmove(fft_data, iq_buffer_pointer + waterfall_offset, c);
          memmove(fft_data + c, iq_buffer_pointer, d);
          waterfall_offset = d;
          waterfall_offset -= WATERFALL_FFT_SIZE; if (waterfall_offset < 0) waterfall_offset += iq_buffer_size;
        } else {
          break;
        };

        MUTEX_UNLOCK(iq_buffer_mutex);

        // Process chunk of data for main waterfall display...

        static float waterfall[WATERFALL_FFT_SIZE] = {};
        static int waterfall_sums = 0;

        ONCE_EVERY_(1 << option_waterfall_smooth) {

          // Turn IQ data into complex floats...

          static double dc_i = 127.5;
          static double dc_q = 127.5;
          //fprintf(stderr, "dc_i = %f, dc_q = %f\n", dc_i, dc_q);
          for (int i = 0; i < WATERFALL_FFT_SIZE; i++) {
            dc_i *= 0.999999; dc_i += 0.000001 * data[2 * i + 0];
            dc_q *= 0.999999; dc_q += 0.000001 * data[2 * i + 1];
            input[i] = ((data[2 * i] - dc_i) + (data[2 * i + 1] - dc_q) * I) / 127.5f;
            input[i] *= (1.0f - cosf(2 * M_PI * i / WATERFALL_FFT_SIZE));
            #if 0
            static double a = 0.0;
            a -= 2 * M_PI * radio_state.offset_frequency / (double) radio_state.sample_rate;
            if (a > M_PI) a -= 2.0 * M_PI;
            if (a < -M_PI) a += 2.0 * M_PI;
            input[i] *= cexpf(a * I);
            #endif
          };
          #if 0
          for (int i = 0; i < WATERFALL_FFT_SIZE; i++) {
            input[i] = 0;
            input[i] += cexpf(2 * M_PI * I * 20 * i / (double) WATERFALL_FFT_SIZE);
            input[i] += cexpf(2 * M_PI * I * 200 * i / (double) WATERFALL_FFT_SIZE);
          };
          #endif

          // Measure headroom for GUI meter...

          for (int i = 0; i < WATERFALL_FFT_SIZE; i++) {
            float now = cabsf(input[i]);
            float t = radio_iq_strength_squared;
            t *= 0.99999; t += 0.00001 * now * now;
            radio_iq_strength_squared = t;
          };

          // Do the FFT...

          fft_fast(&fft_state, output, input);
          for (int i = 0; i < WATERFALL_FFT_SIZE; i++) {
            float value = cabsf(output[(i + (WATERFALL_FFT_SIZE >> 1)) & (WATERFALL_FFT_SIZE - 1)]);
            waterfall[i] += value * value;
            //if (waterfall[i] < value) waterfall[i] = value;
          };
          waterfall_sums++;

          // Scan mode...

          if (radio_scan_enable) {
            float best = 0.0f;
            int best_i = 0;
            for (int i = 0; i < WATERFALL_FFT_SIZE; i++) {
              float value = cabsf(output[(i + (WATERFALL_FFT_SIZE >> 1)) & (WATERFALL_FFT_SIZE - 1)]);
              if (best < value) {
                best = value;
                best_i = i;
              };
            };
            static int current_i = 0;
            float current = cabsf(output[(current_i + (WATERFALL_FFT_SIZE >> 1)) & (WATERFALL_FFT_SIZE - 1)]);
            if (best > current * powf(10.0f, 20.0f / 20.0f)) {
              radio_state.offset_frequency = nearbyint((best_i - 512.0) * IQ_SAMPLE_RATE / WATERFALL_FFT_SIZE);
              current_i = best_i;
              float im = 20.0f * logf(best / current) / logf(10.0f);
              //fprintf(stderr, "Scan mode auto-tuning to %d, best_i=%d, best=%f, current=%f, improvement=%f\n", radio_state.offset_frequency, best_i, best, current, im);
            };
          };

        };

        ONCE_EVERY_(2 * (1 << option_waterfall_slow)) {
          for (int i = 0; i < WATERFALL_FFT_SIZE; i++) {
            waterfall[i] = sqrtf(waterfall[i] / waterfall_sums);
          };
          user_waterfall_append(waterfall);
          memset(waterfall, 0, sizeof(float) * WATERFALL_FFT_SIZE);
          waterfall_sums = 0;
        };

        user_iq_sample(data);

        MUTEX_LOCK(iq_buffer_mutex);
      };

      MUTEX_UNLOCK(iq_buffer_mutex);

    };

    if (!option_less_stuff) {

      // Here's as good a place as any to FFT the audio...

      while (1) {
        MUTEX_LOCK(audio_fft_buffer_mutex);
        if (audio_fft_buffer.size < 512 * sizeof(float)) break;
        float *buffer = (float *) audio_fft_buffer.pointer;
        for (int i = 0; i < 512; i++) {
          input[i] = buffer[i];
          input[i] *= 0.5f * (2.0f - cosf(2 * M_PI * i / 512));
        };
        buffer_free(audio_fft_buffer, 512 * sizeof(float) / 2);
        MUTEX_UNLOCK(audio_fft_buffer_mutex);
        fft_fast(&audio_fft_state, output, input);
        float waterfall[256];
        for (int i = 0; i < 256; i++) {
          waterfall[i] = 2.0f * cabsf(output[i]);
        };
        user_audio_waterfall_append(waterfall);
      };
      MUTEX_UNLOCK(audio_fft_buffer_mutex);

      // ...and the tuned signal...

      while (1) {
        MUTEX_LOCK(signal_fft_buffer_mutex);
        if (signal_fft_buffer.size < 256 * sizeof(complex float)) break;
        complex float *buffer = (complex float *) signal_fft_buffer.pointer;
        for (int i = 0; i < 256; i++) {
          input[i] = buffer[i];
          input[i] *= 0.5f * (2.0f - cosf(2 * M_PI * i / 256));
        };
        buffer_free(signal_fft_buffer, 256 * sizeof(complex float) / 2);
        MUTEX_UNLOCK(signal_fft_buffer_mutex);
        fft_fast(&signal_fft_state, output, input);
        float waterfall[256];
        for (int i = 0; i < 256; i++) {
          waterfall[i] = cabsf(output[(i + 128) & 255]);
        };
        user_signal_waterfall_append(waterfall);
      };
      MUTEX_UNLOCK(signal_fft_buffer_mutex);

    };

  };

  fft_free(&fft_state);
  fft_free(&audio_fft_state);
  fft_free(&signal_fft_state);
  memory_allocate(iq_buffer_pointer, 0);

  thread_run_flag = 0;

  thread_exit();

};

void radio_initialize () {

  radio_state.mode = RADIO_MODE_FM_12;
  radio_state.offset_frequency = 0;
  //radio_state.upconverter_frequency = UPCONVERTER_FREQUENCY;
  radio_state.center_frequency = 155000000;
  radio_state.sample_rate = 2400000;
  radio_state.gain_mode = 1;
  radio_state.direct_sampling = 0;
  radio_state.agc_mode = 0;
  radio_state.gain_index = 28;

  thread_run_flag = 1;
  thread_exit_flag = 0;
  thread_create(thread, NULL);

  while (!initialized) easy_sleep(0.01);

};

void radio_terminate () {

  if (thread_run_flag) {
    thread_exit_flag = 1;
    while (thread_run_flag) easy_sleep(0.01);
  };

};
